'''
Gradient Boosted Trees (GBT) model class for regression and classification tasks.
Supports multiple GBT implementations: scikit-learn, XGBoost, CatBoost, and LightGBM.
'''

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from IPython.display import display
from sklearn.model_selection import train_test_split
from sklearn.model_selection import KFold
from sklearn.ensemble import GradientBoostingClassifier
from xgboost import plot_tree
from xgboost import XGBRegressor, XGBClassifier
from catboost import Pool, CatBoostRegressor, CatBoostClassifier
import lightgbm as lgb
from sklearn.metrics import roc_auc_score, roc_curve, confusion_matrix, precision_recall_curve, f1_score, mean_absolute_error, mean_squared_error
from scipy.stats import kendalltau
import shap
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import ElasticNet
from digital_twinning.utils.gbt_plot_utils import *
from SALib.analyze import sobol


class GBTModel:
    '''
    Gradient Boosted Trees (GBT) model class for regression and classification tasks.
    Supports multiple GBT implementations: scikit-learn, XGBoost, CatBoost, and LightGBM.
    Includes methods for training, prediction, evaluation, and feature importance analysis.
    
    Attributes:
    -----------
    Q : ParameterSet
        Parameter set defining the input features.

    cols : list
        List of feature names.

    label : str
        Name of the quantity of interest (target variable).
    
    method : str
        GBT implementation method ('scikit', 'xgboost', 'catboost', 'lightgbm', 'ElasticNet', 'log_reg').
    
    model_type : str
        Type of model ('regression' or 'classification').
    
    threshold : float or None
        Threshold for classification tasks.
    
    important : list or None
        List of important features to highlight.
    
    model : object
        Trained GBT model.
    
    prediction : function
        Prediction function of the trained model.

    Methods:
    --------
    __init__(self, Q, QoI, ts=None, threshold=None, model_type="regression", gbt_method="xgboost", splitpercent=0.80, splittime=None, split_random=True, important=None)
        Initializes the GBTModel with specified parameters.

    model_dependent_data_preparation(self, d, cols, catcols, method=None)
        Prepares data based on the selected GBT method.

    split_test_and_train_data(self, d, cols, label, threshold, splitpercent, ts=None, splittime=None,
                    resampling=False, split_random=True, verbose_flag=False)
        Splits data into training and testing sets.

    train_model(self, d_X, d_y, n_est=150, max_d=3, learning_rate=0.15, k_fold=None, num_leaves=None, 
                plot_flag=True, ts=None, splitpercent=0.80, splittime=None, split_random=True,
                task_type='GPU', verbose_flag=False)
        Trains the GBT model with optional k-fold cross-validation.

    train_and_validate(self, X_train, y_train, X_val, y_val, num_of_iter=150, max_depth=3, learning_rate=0.15, num_leaves=None, plot_flag=True, task_type='GPU', verbose_flag=True, ts=None, splitpercent=0.80, splittime=None, split_random=True)
        Trains and validates the model on provided datasets.

    build_model(self, X_train, y_train, X_test, y_test, num_iter=150, max_depth=3, learning_rate=0.15, num_leaves=None, plot_flag=True, 
                    task_type='GPU', verbose_flag=False, ts=None, splitpercent=0.80, splittime=None, split_random=True)
        Builds and trains the GBT model.

    predict(self, X_test, **kwargs)
        Makes predictions using the trained model.

    evaluate_model(self, X_train, y_train, X_test, y_test, verbose=False, opt_thresh_comp="from_ROC")
        Evaluates the model's performance.

    compute_optimal_threshold_from_ROC(self, tpr, fpr, thresholds)
        Computes optimal classification threshold from ROC curve.

    compute_optimal_threshold_from_prec_recall(self, precision, recall, thresholds)
        Computes optimal classification threshold from precision-recall curve.

    get_global_feature_importances(self, verbose=True, n_imp_feats=30, feats_to_highlight=None)
        Retrieves global feature importances.

    get_most_important_cat_features(self, imp_feats, n_imp_feats=30, verbose=False)
        Identifies important categorical features.

    print_feature_importances(self, imp, title, feats_to_highlight=None)
        Displays feature importances in a formatted table.

    get_local_feature_importances(self, X_train, q, n=4, plot_flag=False, verbose=False, feats_to_highlight=None)
        Retrieves local feature importances using SHAP values.

    get_shap_values(self, predict_fn, q, forced=False, explainer_type="treeexplainer")
        Computes SHAP values for local interpretability.

    compute_partial_vars(self, model_obj, max_index)
        Computes partial variances for sensitivity analysis.

    get_average_shap_values(self, shap_values)
        Computes average SHAP values for feature importance.

    get_max_shap_values(self, shap_values)
        Computes maximum SHAP values for feature importance.

    get_feats_with_top_n_mixed_impacts(self, n, local_av_imp, local_max_imp)
        Identifies features with top n mixed importances from average and maximum SHAP values.
    
    plot_correlation_between_numeric_features(self, X_train, feats)
        Plots correlation matrix between numeric features.
    
    plot_trees(self, X_train, y_train, n_est)
        Plots the decision trees of the GBT model.
    
    scatterplot_gbt(self, X_train, y_train, feats, threshold, outlier=False, hue=None)
        Plots scatter plots for GBT model features.
    
    get_gbt_model_rules(self, gbt_model, k=None)
        Exports scikit-learn GBT model rules to function.
    
    get_gbt_feature_importance(self, X_train, importances)
        Computes GBT feature importances.
    
    build_lreg_model(self, splitpercent=0.66, splittime=None, verbose=True,
                         split_random=True, solver='newton-cg')
        Builds and evaluates a logistic regression model.
    
    get_lreg_coefficients(lreg_model, cols, regression=False)
        Retrieves coefficients from a logistic regression model.
    
    to_jsonld(self, model_id: str)
        Exports the GBT model metadata to JSON-LD format.
    '''

    def __init__(self, Q, QoI, ts=None, threshold=None, model_type="regression", gbt_method="xgboost", splitpercent=0.80, splittime=None, split_random=True, important=None):  # fmt: skip
        '''
        Initializes the GBTModel with specified parameters.

        Parameters:
        -----------
        Q : VariableSet
            Variable set defining the input features.
        QoI : str
            Name of the quantity of interest (target variable).
        ts : str, optional
            Name of the time series column (default is None).
        threshold : float or None, optional
            Threshold for classification tasks (default is None).
        model_type : str, optional
            Type of model ('regression' or 'classification', default is 'regression').
        gbt_method : str, optional
            GBT implementation method ('scikit', 'xgboost', 'catboost', 'lightgbm', 'ElasticNet', 'log_reg', default is 'xgboost').
        splitpercent : float, optional
            Percentage of data to use for training (default is 0.80).
        splittime : any, optional
            Time value to split data (default is None).
        split_random : bool, optional
            Whether to split data randomly (default is True).
        important : list or None, optional
            List of important features to highlight (default is None).
        '''

        self.Q = Q
        self.cols = self.Q.param_names()
        self.label = QoI
        self.method = gbt_method
        self.model_type = model_type
        self.threshold = threshold
        self.important = important

    def model_dependent_data_preparation(self, d, cols, catcols, method=None):
        '''
        Prepares data based on the selected GBT method.

        Parameters:
        -----------
        d : pd.DataFrame
            Input data.
        cols : list
            List of feature names.
        catcols : list
            List of categorical feature names.
        method : str, optional
            GBT implementation method (default is None, uses self.method).

        Returns:
        -------
        d : pd.DataFrame
            Prepared data.
        cols : list
            Updated list of feature names.
        '''

        if method == None:
            method = self.method

        if method == "scikit" or method == "xgboost":
            # One hot encoding
            # Features to be trained without any categorical ones
            extended_cols = list(set(cols) - set(catcols))
            # One-hot encoding
            if catcols != []:
                ohe = pd.get_dummies(d[catcols])
                extended_cols.extend(list(ohe.columns))
                # Join dataframes
                d = pd.concat([d, ohe], axis=1)
                cols = extended_cols
                # Fill Nans
                d[cols] = d[cols].fillna(0)
                # Change name of categorical features to ohe feature names
                self.catcols = list(ohe.columns)
        elif method == "catboost":
            # convert categories to strings
            for i_col in catcols:
                d[i_col] = d[i_col].astype(str).astype('category')

        return d, cols

    def split_test_and_train_data(self, d, cols, label, threshold, splitpercent, ts=None, splittime=None,
                                  resampling=False, split_random=True, verbose_flag=False):
        '''
        Splits data into training and testing sets.

        Parameters:
        -----------
        d : pd.DataFrame
            Input data.
        cols : list
            List of feature names.
        label : str
            Name of the quantity of interest (target variable).
        threshold : float or None
            Threshold for classification tasks.
        splitpercent : float
            Percentage of data to use for training.
        ts : str, optional
            Name of the time series column (default is None).
        splittime : any, optional
            Time value to split data (default is None).
        resampling : bool, optional
            Whether to apply resampling for class imbalance (default is False).
        split_random : bool, optional
            Whether to split data randomly (default is True).
        verbose_flag : bool, optional
            Whether to print verbose output (default is False).

        Returns:
        -------
        X_train : pd.DataFrame
            Training feature set.
        X_test : pd.DataFrame
            Testing feature set.
        y_train : pd.Series
            Training target variable.
        y_test : pd.Series
            Testing target variable.
        '''

        if split_random == False:
            if splittime == None:
                split_row = int(d.shape[0] * splitpercent)
                splittime = d[ts].iloc[split_row]
            X_train = d[(d[ts] < splittime) & (d[label] == d[label])][cols]
            X_test = d[(d[ts] >= splittime) & (d[label] == d[label])][cols]
            if threshold == None:  # regression
                y_train = d[(d[ts] < splittime) & (d[label] == d[label])][label]
                y_test = d[(d[ts] >= splittime) & (d[label] == d[label])][label]
            else:  # classification
                y_train = d[(d[ts] < splittime) & (d[label] == d[label])][label] > threshold
                y_test = d[(d[ts] >= splittime) & (d[label] == d[label])][label] > threshold
        else:
            X_train, X_test, y_train, y_test = train_test_split(d[cols], d[label], test_size=1 - splitpercent,
                                                                random_state=42)
        return X_train, X_test, y_train, y_test


    def train_model(self, d_X, d_y, n_est=150, max_d=3, learning_rate=0.15, k_fold=None, num_leaves=None, 
                    plot_flag=True, ts=None, splitpercent=0.80, splittime=None, split_random=True,
                    task_type='GPU', verbose_flag=False):
        '''
        Trains the GBT model with optional k-fold cross-validation.

        Parameters:
        -----------
        d_X : pd.DataFrame
            Feature set.
        d_y : pd.DataFrame
            Target variable.
        n_est : int, optional
            Number of estimators (default is 150).
        max_d : int, optional
            Maximum depth of trees (default is 3).
        learning_rate : float, optional
            Learning rate (default is 0.15).
        k_fold : int or None, optional
            Number of folds for cross-validation (default is None).
        num_leaves : int or None, optional
            Number of leaves (default is None).
        plot_flag : bool, optional
            Whether to plot training progress (default is True).
        ts : str, optional
            Name of the time series column (default is None).
        splitpercent : float, optional
            Percentage of data to use for training (default is 0.80).
        splittime : any, optional
            Time value to split data (default is None).
        split_random : bool, optional
            Whether to split data randomly (default is True).
        task_type : str, optional
            Task type for CatBoost ('CPU' or 'GPU', default is 'GPU').
        verbose_flag : bool, optional
            Whether to print verbose output (default is False).

        Returns:
        -------
        model : object
            Trained GBT model.
        '''

        if type(d_X) != pd.DataFrame:
            d_X = pd.DataFrame(d_X)
        if type(d_y) != pd.DataFrame:
            d_y = pd.DataFrame(d_y)
        if k_fold:
            # Perform k-fold cross-validation
            kf = KFold(n_splits=k_fold, shuffle=True, random_state=42)
            fold = 1
            fold_results = []
            for train_indices, val_indices in kf.split(d_X):
                print(f"Training Fold {fold}/{k_fold}...")
                X_train, X_val = d_X.iloc[train_indices], d_X.iloc[val_indices]
                y_train, y_val = d_y.iloc[train_indices], d_y.iloc[val_indices]
                
                # Call build_model for each fold
                model = self.build_model(X_train, y_train, n_est, max_d, learning_rate, 
                                         task_type, verbose_flag)
                
                # Validate the model on the validation set
                predictions = model.predict(X_val)
                if self.model_type == "regression":
                    mse = mean_squared_error(y_val, predictions)
                    print(f"Fold {fold} MSE: {mse:.4f}")
                    fold_results.append(mse)
                # Extend with classification evaluation if needed
                fold += 1
            
            # Report average performance
            print(f"Average MSE across folds: {np.mean(fold_results):.4f}")
        else:
            # Default 80-20 split
            split_index = int(len(d_X) * 0.8)
            X_train, X_val = d_X.iloc[:split_index], d_X.iloc[split_index:]
            y_train, y_val = d_y.iloc[:split_index], d_y.iloc[split_index:]
            
            print("Training with default 80-20 split...")
            model = self.build_model(X_train, y_train, n_est, max_d, learning_rate, num_leaves, plot_flag, 
                    task_type, verbose_flag, ts, splitpercent, splittime, split_random)
            
            predictions = model.predict(X_val)
            if self.model_type == "regression":
                mse = mean_squared_error(y_val, predictions)
                print(f"Validation MSE: {mse:.4f}")
                
        return model
      
    def train_and_validate(self, X_train, y_train, X_val, y_val, num_of_iter=150, max_depth=3, learning_rate=0.15, num_leaves=None, plot_flag=True, task_type='GPU', verbose_flag=True, ts=None, splitpercent=0.80, splittime=None, split_random=True):
        '''
        Trains and validates the model on provided datasets.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.  
        y_train : pd.DataFrame
            Training target variable.
        X_val : pd.DataFrame
            Validation feature set.
        y_val : pd.DataFrame
            Validation target variable.
        num_of_iter : int, optional
            Number of iterations (default is 150).
        max_depth : int, optional
            Maximum depth of trees (default is 3).
        learning_rate : float, optional
            Learning rate (default is 0.15).
        num_leaves : int or None, optional
            Number of leaves (default is None).
        plot_flag : bool, optional
            Whether to plot training progress (default is True).
        task_type : str, optional
            Task type for CatBoost ('CPU' or 'GPU', default is 'GPU').
        verbose_flag : bool, optional
            Whether to print verbose output (default is True).
        ts : str, optional
            Name of the time series column (default is None).
        splitpercent : float, optional
            Percentage of data to use for training (default is 0.80).
        splittime : any, optional
            Time value to split data (default is None).
        split_random : bool, optional
            Whether to split data randomly (default is True).

        Returns:
        -------
        mse_tr : float
            Mean squared error on the training set.
        mse_vl : float
            Mean squared error on the validation set.
        '''

        max_d = max_depth
        n_est = num_of_iter
        
        self.hyper_params = {
            'num_iter': num_of_iter,
            'max_depth': max_d,
            'learning_rate': learning_rate,
            'num_leaves': num_leaves
        }
        
        # Train model
        if self.method == 'scikit':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                model = GradientBoostingRegressor(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                                  max_leaf_nodes=num_leaves)
            elif self.model_type == "classification":  # classification
                model = GradientBoostingClassifier(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                                   max_leaf_nodes=num_leaves)
            # fit model
            model.fit(X_train, y_train)

        elif self.method == 'xgboost':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                if num_leaves == None:
                    model = XGBRegressor(objective="reg:squarederror", learning_rate=learning_rate, n_estimators=n_est,
                                         max_depth=max_d)
                else:
                    model = XGBRegressor(objective="reg:squarederror", learning_rate=learning_rate, n_estimators=n_est,
                                         max_leaves=num_leaves, grow_policy='lossguide')
            elif self.model_type == "classification":  # classification
                if num_leaves == None:
                    model = XGBClassifier(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                          max_leaf_nodes=num_leaves)
                else:
                    model = XGBClassifier(learning_rate=learning_rate, n_estimators=n_est, max_leaf_nodes=num_leaves,
                                          grow_policy='lossguide')
            # fit model
            model.fit(X_train, y_train, verbose=verbose_flag)

        elif self.method == 'catboost':
            cols = list(X_train.columns)
            self.catcols_ind = []
            pool_train = Pool(X_train, y_train, cat_features=self.catcols_ind, feature_names=cols)
            pool_test = Pool(X_val, y_val, cat_features=self.catcols_ind, feature_names=cols)
            # Initialize CatBoostClassifier/Regressor
            if self.model_type == "regression":  # regression
                if num_leaves == None:
                    model = CatBoostRegressor(max_depth=max_d, learning_rate=learning_rate, n_estimators=n_est,
                                              loss_function="RMSE", eval_metric="MAE", use_best_model=True, verbose=0,
                                              task_type=task_type)
                else:
                    model = CatBoostRegressor(learning_rate=learning_rate, grow_policy='Lossguide', max_depth=max_d,
                                              n_estimators=n_est, max_leaves=num_leaves, loss_function="RMSE",
                                              eval_metric="MAE", use_best_model=True, verbose=0, task_type=task_type)
            elif self.model_type == "classification":  # classification
                if num_leaves == None:
                    model = CatBoostClassifier(max_depth=max_d, learning_rate=learning_rate, n_estimators=n_est,
                                               loss_function="Logloss", use_best_model=True, verbose=0,
                                               task_type=task_type)
                else:
                    model = CatBoostClassifier(learning_rate=learning_rate, grow_policy='Lossguide', max_depth=max_d,
                                               n_estimators=n_est, max_leaves=num_leaves, loss_function="Logloss",
                                               use_best_model=True, verbose=0, task_type=task_type)
            # Fit model
            model.fit(pool_train, eval_set=pool_test, plot=plot_flag, verbose=verbose_flag)

        elif self.method == "lightgbm":
            assert not any(' ' in x for x in X_train.columns), 'No space allowed in column names with LightGBM.' 
            pool_train = lgb.Dataset(X_train, label=y_train)  # , categorical_feature=categorical)
            pool_test = lgb.Dataset(X_val, label=y_val)
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                hyper_params = {
                    'objective': 'regression',
                    'metric': 'mse',
                    'boosting': 'gbdt',
                    'max_depth': max_d,
                    'n_estimators': n_est,
                    'num_leaves': num_leaves,
                    'feature_fraction': 0.5,
                    'bagging_fraction': 0.5,
                    'bagging_freq': 20,
                    'learning_rate': learning_rate,
                    'verbose': 0
                }
                self.hyper_params = hyper_params
            elif self.model_type == "classification":  # classification
                hyper_params = {
                    'objective': 'binary',
                    'metric': 'binary_logloss',
                    'num_leaves': num_leaves,
                    'boosting': 'gbdt',
                    'max_depth': max_d,
                    'n_estimators': n_est,
                    'learning_rate': learning_rate,
                    'verbose': 0
                }
                self.hyper_params = hyper_params
            else:
                raise ValueError
            model = lgb.train(hyper_params, pool_train, valid_sets=pool_test, num_boost_round=5000,
                              callbacks=[lgb.early_stopping(stopping_rounds=50)])
        elif self.method == 'ElasticNet':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                model = ElasticNet().fit(X_train, y_train)
            elif self.model_type == "classification":  # classification
                raise ('Use ElasticNet with regression or logistic regression!')
        elif self.method == "log_reg":
            model = self.build_lreg_model()
        else:
            raise ('Unknown method: ', self.method)
        self.model = model   
        self.prediction = model.predict
        y_tr_pred = self.predict(X_train)
        y_vl_pred = self.predict(X_val)
        mse_tr = mean_squared_error(y_train, y_tr_pred)
        mse_vl = mean_squared_error(y_val, y_vl_pred)
        return mse_tr, mse_vl
    
    
    def build_model(self, X_train, y_train, X_test, y_test, num_iter=150, max_depth=3, learning_rate=0.15, num_leaves=None, plot_flag=True, 
                    task_type='GPU', verbose_flag=False, ts=None, splitpercent=0.80, splittime=None, split_random=True):
        '''
        Builds and trains the GBT model.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        y_train : pd.DataFrame
            Training target variable.
        X_test : pd.DataFrame
            Testing feature set.
        y_test : pd.DataFrame
            Testing target variable.
        num_iter : int, optional
            Number of iterations (default is 150).
        max_depth : int, optional
            Maximum depth of trees (default is 3).
        learning_rate : float, optional
            Learning rate (default is 0.15).
        num_leaves : int or None, optional
            Number of leaves (default is None).
        plot_flag : bool, optional
            Whether to plot training progress (default is True).
        task_type : str, optional
            Task type for CatBoost ('CPU' or 'GPU', default is 'GPU').
        verbose_flag : bool, optional
            Whether to print verbose output (default is False).
        ts : str, optional
            Name of the time series column (default is None).
        splitpercent : float, optional
            Percentage of data to use for training (default is 0.80).
        splittime : any, optional
            Time value to split data (default is None).
        split_random : bool, optional
            Whether to split data randomly (default is True).

        Returns:
        -------
        model : object
            Trained GBT model.
        '''

        max_d = max_depth
        n_est = num_iter
        
        # Train model
        if self.method == 'scikit':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                model = GradientBoostingRegressor(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                                  max_leaf_nodes=num_leaves)
            elif self.model_type == "classification":  # classification
                model = GradientBoostingClassifier(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                                   max_leaf_nodes=num_leaves)
            # fit model
            model.fit(X_train, y_train)

        elif self.method == 'xgboost':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                if num_leaves == None:
                    model = XGBRegressor(objective="reg:squarederror", learning_rate=learning_rate, n_estimators=n_est,
                                         max_depth=max_d)
                else:
                    model = XGBRegressor(objective="reg:squarederror", learning_rate=learning_rate, n_estimators=n_est,
                                         max_leaves=num_leaves, grow_policy='lossguide')
            elif self.model_type == "classification":  # classification
                if num_leaves == None:
                    model = XGBClassifier(learning_rate=learning_rate, n_estimators=n_est, max_depth=max_d,
                                          max_leaf_nodes=num_leaves)
                else:
                    model = XGBClassifier(learning_rate=learning_rate, n_estimators=n_est, max_leaf_nodes=num_leaves,
                                          grow_policy='lossguide')
            # fit model
            #model.fit(X_train, y_train, eval_set=[(X_test, y_test)], verbose=verbose_flag)
            model.fit(X_train, y_train, verbose=verbose_flag)

        elif self.method == 'catboost':
            cols = list(X_train.columns)
            self.catcols_ind = []
            pool_train = Pool(X_train, y_train, cat_features=self.catcols_ind, feature_names=cols)
            pool_test = Pool(X_test, y_test, cat_features=self.catcols_ind, feature_names=cols)
            # Initialize CatBoostClassifier/Regressor
            if self.model_type == "regression":  # regression
                if num_leaves == None:
                    model = CatBoostRegressor(max_depth=max_d, learning_rate=learning_rate, n_estimators=n_est,
                                              loss_function="RMSE", eval_metric="MAE", use_best_model=True, verbose=0,
                                              task_type=task_type)
                else:
                    model = CatBoostRegressor(learning_rate=learning_rate, grow_policy='Lossguide', max_depth=max_d,
                                              n_estimators=n_est, max_leaves=num_leaves, loss_function="RMSE",
                                              eval_metric="MAE", use_best_model=True, verbose=0, task_type=task_type)
            elif self.model_type == "classification":  # classification
                if num_leaves == None:
                    model = CatBoostClassifier(max_depth=max_d, learning_rate=learning_rate, n_estimators=n_est,
                                               loss_function="Logloss", use_best_model=True, verbose=0,
                                               task_type=task_type)
                else:
                    model = CatBoostClassifier(learning_rate=learning_rate, grow_policy='Lossguide', max_depth=max_d,
                                               n_estimators=n_est, max_leaves=num_leaves, loss_function="Logloss",
                                               use_best_model=True, verbose=0, task_type=task_type)
            # Fit model
            model.fit(pool_train, eval_set=pool_test, plot=plot_flag, verbose=verbose_flag)

        elif self.method == "lightgbm":
            assert not any(' ' in x for x in X_train.columns), 'No space allowed in column names with LightGBM.' 
            pool_train = lgb.Dataset(X_train, label=y_train)  # , categorical_feature=categorical)
            pool_test = lgb.Dataset(X_test, label=y_test)
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                hyper_params = {
                    'objective': 'regression',
                    'metric': 'mse',
                    'boosting': 'gbdt',
                    'max_depth': max_d,
                    'n_estimators': n_est,
                    'num_leaves': num_leaves,
                    'feature_fraction': 0.5,
                    'bagging_fraction': 0.5,
                    'bagging_freq': 20,
                    'learning_rate': learning_rate,
                    'verbose': 0
                }
            elif self.model_type == "classification":  # classification
                hyper_params = {
                    'objective': 'binary',
                    'metric': 'binary_logloss',
                    'num_leaves': num_leaves,
                    'boosting': 'gbdt',
                    'max_depth': max_d,
                    'n_estimators': n_est,
                    'learning_rate': learning_rate,
                    'verbose': 0
                }
            else:
                raise ValueError
            model = lgb.train(hyper_params, pool_train, valid_sets=pool_test, num_boost_round=5000,
                              callbacks=[lgb.early_stopping(stopping_rounds=50)])
        elif self.method == 'ElasticNet':
            # Initialize Classifier/Regressor
            if self.model_type == "regression":  # regression
                model = ElasticNet().fit(X_train, y_train)
            elif self.model_type == "classification":  # classification
                raise ('Use ElasticNet with regression or logistic regression!')
        elif self.method == "log_reg":
            model = self.build_lreg_model()
        else:
            raise ('Unknown method: ', self.method)
        self.model = model
        return model

    def predict(self, X_test, **kwargs):
        '''
        Makes predictions using the trained model.

        Parameters:
        -----------
        X_test : pd.DataFrame
            Testing feature set.
        **kwargs : dict
            Additional keyword arguments for the prediction function.

        Returns:
        -------
        pred : np.ndarray
            Predicted values.
        '''

        pred = self.prediction(X_test)
        if pred.ndim == 1:
            pred = pred.reshape(-1, 1)
        return pred
    
    def evaluate_model(self, X_train, y_train, X_test, y_test, verbose=False, opt_thresh_comp="from_ROC"):
        '''
        Evaluates the model's performance.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        y_train : pd.DataFrame
            Training target variable.
        X_test : pd.DataFrame
            Testing feature set.
        y_test : pd.DataFrame
            Testing target variable.
        verbose : bool, optional
            Whether to print verbose output (default is False).
        opt_thresh_comp : str, optional
            Method to compute optimal threshold for classification ('from_ROC' or 'from_prec', default is 'from_ROC').

        Returns:
        -------
        results : list
            List containing a DataFrame of evaluation metrics and a dictionary of model evaluation results.
        '''

        if self.model_type == "classification":  # For classification problem

            if self.method == "lightgbm":
                pred = self.model.predict(X_test)
            else:
                pred = self.model.predict_proba(X_test)[:, 1]

            # Get optimal value of threshold
            fpr, tpr, thresh_ROC = roc_curve(y_test, pred)
            precision, recall, thresh_prec = precision_recall_curve(y_test, pred)

            if opt_thresh_comp == "from_ROC":
                opt_thresh = self.compute_optimal_threshold_from_ROC(tpr, fpr, thresh_ROC)
            elif opt_thresh_comp == "from_prec":
                opt_thresh = self.compute_optimal_threshold_from_prec_recall(precision, recall, thresh_prec)
            else:
                raise ("There is no optimal threshold method {}. Possible options are from_ROC or from_prec".format(
                    opt_thresh_comp))

            # Get predicted classes

            pred_class = pred > opt_thresh

            # if self.method == "catboost":
            #    pred_class = pred_class == "True"
            # evaluate model
            model_eval = {
                "auc": roc_auc_score(y_test, pred),
                "f1_score": f1_score(np.array(y_test), pred_class)
            }

            if verbose:
                plot_ROC_and_recall_curve = True
                plot_conf_matrix = True
                print(model_eval)
            else:
                plot_ROC_and_recall_curve = False
                plot_conf_matrix = False

            if plot_ROC_and_recall_curve:
                # Plot ROC CURVE
                plt.figure(figsize=(10, 3))
                plt.subplot(1, 2, 1)
                plt.plot(fpr, tpr)
                plt.title("ROC curve")
                plt.xlabel("False Positive Rate")
                plt.ylabel("True Positive Rate")
                # Plot recall curve
                plt.subplot(1, 2, 2)
                plt.plot(recall, precision)
                plt.title("Precision-recall curve")
                plt.xlabel("Recall")
                plt.ylabel("Precision")
                plt.show()

            if plot_conf_matrix:
                # Print confusion matrix
                print("Confusion Matrix:")
                conf_matrix = pd.DataFrame(confusion_matrix(y_test, pred_class),
                                           columns=["predicted GOOD quality", "predicted BAD quality"],
                                           index=["actual GOOD quality", "actual BAD quality"])
                display(conf_matrix)


        elif self.model_type == "regression":  # for regression problem
            pred = self.model.predict(X_test)
            model_eval = {
                "Kendall_tau": kendalltau(y_test, pred)[0],
                "MSE": mean_squared_error(y_test, pred),
                "MAE": mean_absolute_error(y_test, pred),
                "RMSE": np.sqrt(mean_squared_error(y_test, pred)),
                "STD_of_label": self.y_train.std()
            }
            if verbose == True:
                print('Kendall tau correlation - measure of correspondance between two rankings: %.3f' % model_eval["Kendall_tau"])
                #print('Pearson correlation - measure of linear realationship (cov normalised): %.3f' % model_eval["Pearson"])
                #print('Spearman correaltion - cov(rank(y1), rank(y2)/stdv(rank(y1))): %.3f' % model_eval["Spearman"])
                print("mean squared error: ", model_eval["MSE"])
                print("Root mean squared error: ", model_eval["RMSE"]) 
                print("STD of label:", model_eval["STD_of_label"])
                print("MAE: ", model_eval["MAE"])
        # pdb.set_trace()
        df = pd.DataFrame(model_eval, index=[0])
        df['X_train_cols'] = [X_train.columns]
        df['label'] = [y_train.columns] #name
        return [df, model_eval]

    def compute_optimal_threshold_from_ROC(self, tpr, fpr, thresh):
        '''
        Computes the optimal threshold from ROC curve.

        Parameters:
        -----------
        tpr : np.ndarray
            True positive rates.
        fpr : np.ndarray
            False positive rates.
        thresh : np.ndarray
            Thresholds.

        Returns:
        -------
        ROC_thresh : float
            Optimal threshold based on ROC curve.    
        '''

        J = tpr - fpr
        ix = np.argmax(J)
        ROC_thresh = thresh[ix]
        return ROC_thresh

    def compute_optimal_threshold_from_prec_recall(self, precision, recall, thresh):
        '''
        Computes the optimal threshold from precision-recall curve.

        Parameters:
        -----------
        precision : np.ndarray
            Precision values.
        recall : np.ndarray
            Recall values.
        thresh : np.ndarray
            Thresholds.

        Returns:
        -------
        prec_thresh : float
            Optimal threshold based on precision-recall curve.
        '''

        # convert to f score
        fscore = (2 * precision * recall) / (precision + recall)
        # locate the index of the largest f score
        ix = np.argmax(fscore)
        prec_thresh = thresh[ix]
        return prec_thresh

    def get_global_feature_importances(self, verbose=True, n_imp_feats=30, feats_to_highlight=None):
        '''
        Computes and returns global feature importances.

        Parameters:
        -----------
        verbose : bool, optional
            Whether to print verbose output (default is True).
        n_imp_feats : int, optional
            Number of top important features to consider (default is 30).
        feats_to_highlight : list or str or None, optional
            Features to highlight in the output (default is None).

        Returns:
        -------
        global_imp : pd.Series
            Series containing global feature importances.
        imp_feats : list
            List of important features.
        imp_cat_feats : list
            List of important categorical features.
        '''

        # Pandas series with important features and their importances
        if self.method == "lightgbm":
            global_imp = self.get_gbt_feature_importance(self.model.feature_importance(importance_type='gain'))
        else:
            global_imp = self.get_gbt_feature_importance(self.model.feature_importances_)
        # Important features
        imp_feats = global_imp.index
        # Get most important categorical features
        imp_cat_feats = self.get_most_important_cat_features(imp_feats, n_imp_feats=n_imp_feats, verbose=verbose)
        # Print global importance list
        if verbose:
            self.print_feature_importances(global_imp, "Global feature importances",
                                           feats_to_highlight=feats_to_highlight)
        return global_imp, imp_feats, imp_cat_feats

    def get_most_important_cat_features(self, imp_feats, n_imp_feats=30, verbose=False):
        '''
        Identifies the most important categorical features.

        Parameters:
        -----------
        imp_feats : list
            List of important features.
        n_imp_feats : int, optional
            Number of top important features to consider (default is 30).
        verbose : bool, optional
            Whether to print verbose output (default is False).

        Returns:
        -------
        imp_cat_feats : list
            List of important categorical features.
        '''

        imp_cat_feats = list(set(imp_feats[:n_imp_feats]).intersection(set(self.catcols)))
        if verbose:
            if (len(imp_cat_feats) > 0):
                print("Categorical features in the top {} importance list:".format(n_imp_feats))
                print(imp_cat_feats)
            else:
                print("There is no categorical feature in the top {} importance list".format(n_imp_feats))

        return imp_cat_feats

    def print_feature_importances(self, imp, title, feats_to_highlight=None):
        '''
        Prints feature importances in a formatted table.

        Parameters:
        -----------
        imp : pd.Series
            Series containing feature importances.
        title : str
            Title for the importance table.
        feats_to_highlight : list or str or None, optional
            Features to highlight in the output (default is None).

        Returns:
        -------
        None
        '''
        print("")
        print(title)
        imp_df = pd.DataFrame()
        imp_df["feature"] = imp.index
        imp_df["importance"] = imp.values
        pd.options.display.max_colwidth = 100
        pd.options.display.max_rows = 200
        if feats_to_highlight == None:
            imp_form = imp_df.style.apply(
                lambda x: ['background-color: lightblue' if x.feature in self.catcols else '' for i in x], axis=1)
            display(imp_form)
        elif isinstance(feats_to_highlight, str):
            imp_form = imp_df.style.apply(lambda x: [
                'background-color: lightblue' if x.feature in self.catcols else 'background-color: orange' if (
                            x.feature in feats_to_highlight) else "" for i in x], axis=1)
            display(imp_form)
        else:  # if feats_to_highlight is a list
            imp_form = imp_df.style.apply(lambda x: ['background-color: orange' if any([x.feature in f for f in
                                                                                        feats_to_highlight]) else 'background-color: lightblue' if x.feature in self.catcols else ""
                                                     for i in x], axis=1)
            display(imp_form)

    def get_local_feature_importances(self, X_train, q, n=4, plot_flag=False, verbose=False, feats_to_highlight=None):
        '''
        Computes and returns local feature importances using SHAP values.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        q : pd.DataFrame
            Query points for which local feature importances are computed.
        n : int, optional
            Number of top important features to consider (default is 4).
        plot_flag : bool, optional
            Whether to plot SHAP summary plots (default is False).
        verbose : bool, optional
            Whether to print verbose output (default is False).
        feats_to_highlight : list or str or None, optional
            Features to highlight in the output (default is None).

        Returns:
        -------
        shap_values : np.ndarray
            Array of SHAP values.
        local_av_imp : pd.Series
            Series containing average local feature importances.
        local_max_imp : pd.Series
            Series containing maximum local feature importances.
        feats_with_top_n_mixed_imp : list
            List of features with top n mixed importances.                
        '''

        shap.initjs()
        explainer = shap.TreeExplainer(self.model)
        shap_values = explainer.shap_values(q)
        if isinstance(shap_values, list) and len(
                shap_values) == 2:  # if shap gives an array for both classes (True and False)
            shap_values = shap_values[1]
        local_av_imp = self.get_average_shap_values(shap_values)
        local_max_imp = self.get_max_shap_values(shap_values)
        feats_with_top_n_mixed_imp = self.get_feats_with_top_n_mixed_impacts(n, local_av_imp, local_max_imp)
        if plot_flag:
            # Plot summary of average effect of the features
            shap.summary_plot(shap_values, q, plot_type="bar", plot_flag=True, matplotlib=True)
            # Plot summary of local effect of the features
            shap.summary_plot(shap_values, X_train, plot_flag=True)
            plt.show()
        if verbose:
            self.print_feature_importances(local_av_imp, "Average local importances (average SHAP values)",
                                           feats_to_highlight=feats_to_highlight)
            self.print_feature_importances(local_max_imp, "Maximum local importances (max SHAP values)",
                                           feats_to_highlight=feats_to_highlight)
            print("Features with top {} highest local and average impact:".format(n))
            print(feats_with_top_n_mixed_imp)
        return shap_values, local_av_imp, local_max_imp, feats_with_top_n_mixed_imp
    
    def get_shap_values(self, q, forced=False, explainer_type="treeexplainer"):
        '''
        Computes SHAP values for the given query points.

        Parameters:
        -----------
        q : pd.DataFrame
            Query points for which SHAP values are computed.
        forced : bool, optional
            Whether to force re-computation of the explainer (default is False).
        explainer_type : str, optional
            Type of SHAP explainer to use (default is "treeexplainer").

        Returns:
        -------
        shap_values : np.ndarray
            Array of SHAP values.
        '''

        if explainer_type == "treeexplainer":
            if hasattr(self, 'explainer') == False or forced == True:
                explainer = shap.TreeExplainer(self.model)
                self.explainer = explainer
        shap_values = self.explainer(q)
        return shap_values
    
    def compute_partial_vars(self, model_obj, max_index):
        '''
        Computes partial variances using Sobol sensitivity analysis.
        
        Parameters:
        -----------
        model_obj : object
            Model object containing parameters and prediction methods.
        max_index : int
            Maximum order of Sobol indices to compute (1 or 2).

        Returns:
        -------
        partial_var_df : pd.DataFrame
            DataFrame containing partial variances.
        sobol_index_df : pd.DataFrame
            DataFrame containing Sobol indices.
        y_var : np.ndarray
            Array of variances of the model outputs.
        '''

        paramset = model_obj.Q
        QoI_names = model_obj.QoI_names
        problem = {
            'num_vars': paramset.num_params(), 'names': paramset.param_names(), 'dists': paramset.get_dist_types(), 'bounds': paramset.get_dist_params()
            }
        q = paramset.sample(method='Sobol_saltelli', n=32768) # saltelli working only for uniform distribution
        # https://salib.readthedocs.io/en/latest/user_guide/advanced.html
        q_df = pd.DataFrame(q, columns=paramset.param_names())
        y = model_obj.predict(q_df)
        
        # Run model
        S1 = []
        S2 = []
        for i in range(y.shape[1]):
            y_i = y[:,i]

            # Sobol analysis
            Si_i = sobol.analyze(problem, y_i)
            T_Si, first_Si, (idx, second_Si) = sobol.Si_to_pandas_dict(Si_i)
            df = Si_i.to_df()
            cols_S1 = list(df[1].index)
            cols_S2 = list(df[2].index)

            S1.append(first_Si['S1'])
            S2.append(second_Si['S2'])

        S1 = np.array(S1)
        S2 = np.array(S2)

        col_names = cols_S1
        sobol_index = S1
        if max_index == 2:
            sobol_index = np.concatenate([S1, S2], axis=1)
            col_names = cols_S1 + cols_S2
            col_names = [f"{x[0]} {x[1]}" if isinstance(x, tuple) else x for x in col_names]
        
        y_var = np.broadcast_to(y.var(axis=0).reshape(-1,1), sobol_index.shape)
        partial_variance = np.multiply(sobol_index, y_var)
        
        y_var = y.var(axis=0).reshape(-1, 1)
        partial_variance = sobol_index * y_var
             
        partial_var_df, sobol_index_df = pd.DataFrame(partial_variance, columns=col_names, index=QoI_names), pd.DataFrame(sobol_index, columns=col_names, index=QoI_names)
        
        return partial_var_df, sobol_index_df, y_var

    def get_average_shap_values(self, shap_values):
        '''
        Computes average SHAP values for feature importance.

        Parameters:
        -----------
        shap_values : np.ndarray
            Array of SHAP values.

        Returns:
        -------
        local_av_imp : pd.Series
            Series containing average local feature importances.
        '''

        av_shap_values = np.mean(np.abs(shap_values), axis=0)
        local_av_imp = self.get_gbt_feature_importance(av_shap_values)
        return local_av_imp

    def get_max_shap_values(self, shap_values):
        '''
        Computes maximum SHAP values for feature importance.

        Parameters:
        -----------
        shap_values : np.ndarray
            Array of SHAP values.

        Returns:
        -------
        local_max_imp : pd.Series
            Series containing maximum local feature importances.
        '''

        av_shap_values = np.max(np.abs(shap_values), axis=0)
        local_max_imp = self.get_gbt_feature_importance(av_shap_values)
        return local_max_imp

    def get_feats_with_top_n_mixed_impacts(self, n, local_av_imp, local_max_imp):
        '''
        Identifies features with top n mixed importances from average and maximum SHAP values.

        Parameters:
        -----------
        n : int
            Number of top important features to consider.
        local_av_imp : pd.Series
            Series containing average local feature importances.
        local_max_imp : pd.Series
            Series containing maximum local feature importances.

        Returns:
        -------
        feats_with_top_n_mixed_imp : list
            List of features with top n mixed importances.
        '''

        top_n_av = list(local_av_imp.iloc[:n].index)
        top_n_max = list(local_max_imp.iloc[:n].index)
        feats_with_top_n_mixed_imp = list((set(top_n_av)).union(set(top_n_max)))
        return feats_with_top_n_mixed_imp

    def plot_correlation_between_numeric_features(self, X_train, feats):
        '''
        Plots correlation matrix between numeric features.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        feats : list
            List of features to consider for correlation plot.
        
        Returns:
        -------
        None
        '''

        feats = set(feats) - set(self.catcols)
        d = pd.concat([X_train[feats], self.y_train], axis=1)
        corrplot(df=d[feats])

    def plot_trees(self, X_train, y_train, n_est):
        '''
        Plots the decision trees of the GBT model.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        y_train : pd.DataFrame
            Training target variable.
        n_est : int
            Number of estimators (trees) to plot.

        Returns:
        -------
        None
        '''

        cols = list(X_train.columns)
        # n_est
        if self.method == 'catboost':
            pool_train = Pool(X_train, y_train, cat_features=self.catcols_ind, feature_names=cols)
            for i in range(n_est):
                print('Tree', i, ':')
                print(self.model.plot_tree(tree_idx=i, pool=pool_train))
        elif self.method == "scikit":
            self.get_gbt_model_rules(self.model)
        elif self.method == "lightgbm":
            # fig, ax = plt.subplots(n_est, figsize=(10, 30))
            for i in range(n_est):
                # lgb.create_tree_digraph(self.model, tree_index=i)
                lgb.plot_tree(self.model, tree_index=i)
        elif self.method =="xgboost":
            fig, ax = plt.subplots(n_est,figsize=(30, 30))
            for i in range(n_est):
                plot_tree(self.model, num_trees=i, ax=ax[i])
            plt.show()

    def scatterplot_gbt(self, X_train, y_train, feats, threshold, outlier=False):
        '''
        Plots scatter plots for GBT model features.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        y_train : pd.DataFrame
            Training target variable.
        feats : list
            List of features to plot.
        threshold : float
            Threshold value for outlier detection.
        outlier : bool, optional
            Whether to highlight outliers (default is False).

        Returns:
        -------
        None
        '''

        d = pd.concat([X_train, y_train], axis=1)
        label = self.y_train.name
        for f in feats:
            scatterplot(d, label, f, outlier, threshold=threshold)
            scatterplot(d, label, f, outlier, threshold=threshold, regressionline=True)


    def get_gbt_model_rules(self, gbt_model, k=None):
        '''
        Exports scikit-learn GBT model rules to function.

        Parameters:
        -----------
        gbt_model : object
            Trained GBT model.
        k : int or None, optional
            Number of trees to consider (default is None, which means all trees).

        Returns:
        -------
        func_str : list
            List of strings representing the function to score using GBT rules.
        '''

        if len(self.cols) == 1:
            # duplication of the same feature name is needed if only 1 was used for GBT training!
            column_names = self.cols * 3
        INDENT = str("    ")
        num_of_trees = len(gbt_model.estimators_)
        if k != None:
            num_of_trees = min(k, num_of_trees)

        func_str = ["def score_gbt(eval_df, model_name):"]
        func_str += [INDENT + '"""Score segment by gbt rule"""']
        func_str += []
        func_str += [INDENT + "def score_by_gbt_tree_rule(row):"]
        func_str += [INDENT + '# GBT model generated by Scikit-Learn']
        func_str += [INDENT + INDENT + "score = 0.0", ]
        for i in range(num_of_trees):
            func_str += [INDENT + INDENT + '### tree_%i ###' % (i + 1)]
            func_str += print_tree_with_names(gbt_model.estimators_[i, 0].tree_, self.cols, INDENT)
            func_str += ['']
        func_str += [INDENT + INDENT + "return score"]
        func_str += []
        func_str += [INDENT + "eval_df['SCORE_'+model_name] = eval_df.apply(score_by_gbt_tree_rule, axis=1)"]
        func_str += [INDENT + "return eval_df"]
        for line in func_str:
            print(line)
        return func_str

    def get_gbt_feature_importance(self, X_train, importances):
        '''
        Computes GBT feature importances.

        Parameters:
        -----------
        X_train : pd.DataFrame
            Training feature set.
        importances : np.ndarray
            Array of feature importances.

        Returns:
        -------
        gbt_all_importances : pd.Series
            Series containing GBT feature importances.
        '''

        gbt_all_importances = pd.Series(importances, index=X_train.columns, name="feature importance").sort_values(
            ascending=False)
        return gbt_all_importances[gbt_all_importances > 0]

    def build_lreg_model(self, splitpercent=0.66, splittime=None, verbose=True,
                         split_random=True, solver='newton-cg'):
        '''
        Builds and evaluates a logistic regression model.

        Parameters:
        -----------
        splitpercent : float, optional
            Percentage of data to use for training (default is 0.66).
        splittime : any, optional
            Time value to split data (default is None).
        verbose : bool, optional
            Whether to print verbose output (default is True).
        split_random : bool, optional
            Whether to split data randomly (default is True).
        solver : str, optional
            Solver to use for logistic regression (default is 'newton-cg').
            
        Returns:
        -------
        auc : float
            Area Under the Curve (AUC) score of the logistic regression model.
        '''

        d = self.df[self.important + [self.label, 'ts']].fillna(0)

        if split_random == False:
            if splittime == None:
                split_row = int(d.shape[0] * splitpercent)
                splittime = d["ts"].iloc[split_row]
            features_train = d[(d['ts'] < splittime) & (d[self.label] == d[self.label])][self.important]
            features_test = d[(d['ts'] >= splittime) & (d[self.label] == d[self.label])][self.important]
            if self.threshold == None:  # regression
                labels_train = d[(d['ts'] < splittime) & (d[self.label] == d[self.label])][self.label]
                labels_test = d[(d['ts'] >= splittime) & (d[self.label] == d[self.label])][self.label]
            else:  # classification
                labels_train = d[(d['ts'] < splittime) & (d[self.label] == d[self.label])][self.label] > self.threshold
                labels_test = d[(d['ts'] >= splittime) & (d[self.label] == d[self.label])][self.label] > self.threshold
        else:
            features_train, features_test, labels_train, labels_test = train_test_split(d[self.important], d[self.label],
                                                                                        test_size=1 - splitpercent,
                                                                                        random_state=42)

        logreg = LogisticRegression(solver=solver)

        logreg.fit(features_train, labels_train)
        pred = logreg.predict_proba(features_test)[:, 1]
        auc = roc_auc_score(labels_test, pred)
        print('Logreg model for', self.label, 'AUC =', auc)
        if verbose:
            pretty_print_df(df=self.get_lreg_coefficients(logreg, self.important))
        return auc

    @staticmethod
    def get_lreg_coefficients(lreg_model, cols, regression=False):
        '''
        Retrieves coefficients from a logistic regression model.

        Parameters:
        -----------
        lreg_model : object
            Trained logistic regression model.
        cols : list
            List of feature names.
        regression : bool, optional
            Whether the model is for regression (default is False).

        Returns:
        -------
        coef_df : pd.DataFrame
            DataFrame containing feature names and their coefficients.
        '''

        intercept_df = pd.DataFrame({"name": "(intercept)", "value": lreg_model.intercept_})
        if regression:
            coef_df = pd.DataFrame({"name": cols, "value": lreg_model.coef_}).sort_values(by='value')
        else:
            coef_df = pd.DataFrame({"name": cols, "value": lreg_model.coef_[0]}).sort_values(by='value')
        return intercept_df.append(coef_df, ignore_index=True)

    def to_jsonld(self, model_id: str):
        '''
        Exports the GBT model metadata to JSON-LD format.
        
        Parameters:
        -----------
        model_id : str
            Unique identifier for the model.

        Returns:
        -------
        jsonld : dict
            JSON-LD representation of the model metadata.
        '''

        match self.method:
            case 'scikit':
                algorithm = 'https://en.wikipedia.org/wiki/Gradient_boosting'
            case 'xgboost':
                algorithm = 'https://en.wikipedia.org/wiki/XGBoost'
            case 'catboost':
                algorithm = 'https://en.wikipedia.org/wiki/CatBoost'
            case 'lightgbm':
                algorithm = 'https://en.wikipedia.org/wiki/LightGBM'
            case 'ElasticNet':
                algorithm = 'https://en.wikipedia.org/wiki/Elastic_net_regularization'
            case 'log_reg':
                algorithm = 'https://en.wikipedia.org/wiki/Logistic_regression'

        jsonld = {

            "@context": {
                "mls": "https://ml-schema.github.io/documentation/mls.html",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#"
            },

            "@id": f"https://example.org/models/{model_id}",
            "@type": "mls:Model",
            "mls:implementsAlgorithm": {
                "@id": algorithm,
                "@type": "mls:Algorithm",
                "rdfs:label": self.method,
            },

            "mls:hasHyperParameter": [
                {
                    "@type": "mls:HyperParameterSetting",
                    "mls:hasParameterName": name,
                    "mls:hasParameterValue": str(value).lower() if isinstance(value, bool) else str(value)
                }
                for name, value in self.hyper_params.items()

            ],

            "mls:hasInput": [
                {
                    "@type": "mls:Feature",
                    "mls:featureName": name,
                    "mls:hasDistribution": {
                        "@type": "mls:Distribution",
                        "mls:distributionType": dist.get_type(),
                        "mls:params": str(dist.dist_params),
                    }
                }
                for (name, dist) in self.Q.params.items()
            ]
        }
    
        return jsonld


from setuptools import setup, find_packages
from pathlib import Path

setup(
    name='digital_twinning',
    version='0.1.0',
    author='András Urbanics, Áron Friedman, Bence Popovics, Emese Vastag, Noémi Friedman',
    author_email='popbence@hun-ren.sztaki.hu',
    description='A comprehensive package for digital twin model updating and predictive modeling using machine learning and uncertainty quantification techniques',
    long_description = Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type='text/markdown',
    url='https://github.com/TRACE-Structures/digital_twinning/',
    packages=find_packages(),
    py_modules=['digital_twinning'],
    install_requires=[
        'asyncio',
        'catboost',
        'emcee',
        'gPCE_model',
        'IPython',
        'lightgbm',
        'matplotlib',
        'numpy',
        'pandas',
        'plotly',
        'SALib',
        'scipy',
        'seaborn',
        'shap',
        'scikit-learn',
        'torch',
        'uncertain_variables',
        'xgboost'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
    ],
)



""" JointManager for multi-building model updating using MCMC"""

import uncertain_variables as uv
import pandas as pd
import emcee
import time
from digital_twinning.utils import utils
import numpy as np

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)


class JointManager:
    """ Class to manage joint model updating for multiple buildings through one or more shared parameters
    
        Attributes
        ----------
        models : list of SurrogateModel
            List of surrogate models for each building
            
        joint_parameters : dict
            Dictionary defining joint parameters across buildings
            
        Q : VariableSet
            Joint VariableSet for all buildings
            
        indices : list of list of int
            Indices mapping building parameters to joint parameters
            
        scale : list of list of float
            Scaling factors for each building's parameters
            
        shift : list of list of float
            Shift values for each building's parameters
            
        Methods
        -------
        __init__(models, joint_parameters)
            Initialize the JointManager with given models and joint parameters
            
        generate_joint_stdrn_simparamset(sigmas)
            Generate joint standard random simulation variable set
            
        choose_distribution(buildings, joint_names, mode='auto')
            Choose distribution for joint parameters
            
        get_joint_paramset_and_indices(buildings, joint_parameters)
            Create joint VariableSet and mapping indices
            
        get_logprob(q)
            Compute log-probability for MCMC
            
        logprior(q)
            Compute log-prior probability
            
        loglikelihood(q, y_m)
            Compute log-likelihood of observed data
            
        update(y_list, sigmas, nwalkers=150, nburn=100, niter=350, Q_='default', plot_samples=False)
            Perform MCMC update with observed data
            
        get_mean_and_var_by_model(scaled=True)
            Get mean and variance of posterior for each building
            
        get_MAP_by_model(scaled=True)
            Get maximum a posteriori estimate for each building
            
        get_MAP()
            Get maximum a posteriori estimate for joint parameters
            
        get_data_of_model(model_index)
            Get MAP, mean, variance, and samples for a specific building
            
        get_posterior_samples_by_model(scaled=True)
            Get posterior samples for each building
            
        get_posterior_samples()
            Get posterior samples for joint parameters """
    
    def __init__(self, models, joint_parameters):
        """ Initialize the JointManager with given models and joint parameters

            Parameters
            ----------
            models : list of SurrogateModel
                List of surrogate models for each building

            joint_parameters : dict
                Dictionary defining joint parameters across buildings """
        
        self.models = models
        self.joint_parameters = joint_parameters
        self.Q, self.indices, self.scale, self.shift = self.get_joint_paramset_and_indices(models, joint_parameters)

    def generate_joint_stdrn_simparamset(self, sigmas):
        """ Generate joint standard random simulation variable set

            Parameters
            ----------
            sigmas : list of pandas.DataFrame
                List of standard deviation DataFrames for each building
            
            Returns
            -------
            E_joint : VariableSet
                Joint standard random simulation variable set """
        
        joint_sigma = np.zeros((1, 0))
        for i in range(len(sigmas)):
            building_sigma = sigmas[i].values
            joint_sigma = np.concatenate((joint_sigma, building_sigma), axis=1)
        joint_sigma = joint_sigma.reshape(-1, 1)
        E_joint = utils.generate_stdrn_simparamset(joint_sigma)
        return E_joint
    
    def choose_distribution(self, buildings, joint_names, mode='auto'):
        """ Choose distribution for joint parameters

            Parameters
            ----------
            buildings : list of SurrogateModel
                List of surrogate models of the buildings

            joint_names : list of str
                List of joint parameter names for each building

            mode : str, optional
                Mode for choosing distribution ('first_dist', 'uniform', 'normal', 'auto'), by default 'auto'

            Returns
            -------
            distribution : Distribution
                Chosen distribution for the joint parameter"""
        
        if mode == 'first_dist':
            for i in range(len(joint_names)):
                if joint_names[i] != '-':
                    break
            distribution = buildings[i].Q.params[joint_names[i]]
            return distribution
        
        elif mode == 'uniform':
            distribution = uv.UniformDistribution(0, 1)
            return distribution
        
        elif mode == 'normal':
            distribution = uv.NormalDistribution(0, 1)
            return distribution
        
        elif mode == 'auto':
            for i in range(len(joint_names)):
                if joint_names[i] != '-':
                    break
            if buildings[i].Q.params[joint_names[i]].get_dist_type() == 'unif':
                distribution = uv.UniformDistribution(0, 1)
                return distribution
            elif buildings[i].Q.params[joint_names[i]].get_dist_type() == 'norm':
                distribution = uv.NormalDistribution(0, 1)
                return distribution                
    
    def get_joint_paramset_and_indices(self, buildings, joint_parameters):
        """ Create joint VariableSet and mapping indices

            Parameters
            ----------
            buildings : list of SurrogateModel
                List of surrogate models of the buildings

            joint_parameters : dict
                Dictionary defining joint parameters across buildings

            Returns
            -------
            Q_joint : VariableSet
                Joint VariableSet for all buildings

            indices : list of list of int
                Indices mapping building parameters to joint parameters

            scale : list of list of float
                Scaling factors for each building's parameters

            shift : list of list of float
                Shift values for each building's parameters """
        
        building_parameters = []
        for b in range(len(buildings)):
            building_parameters.append(buildings[b].Q.param_names())
        joint_names = list(joint_parameters.keys())
        for i in range(len(joint_parameters)):
            parameter_list = joint_parameters[joint_names[i]]
            assert len(parameter_list) == len(building_parameters), "The jointparameters should contain every building. Put a '-' when you don't want to use the building for parameter join"
            for j in range(len(building_parameters)):
                if parameter_list[j] != '-':
                    assert parameter_list[j] in building_parameters[j], "Parameter '{}' is not in parameter list {}".format(parameter_list[j], building_parameters[j])

        Q_joint = uv.VariableSet()

        indices = []
        scale = []
        shift = []
        for i in range(len(building_parameters)):
            indices.append([None]*len(building_parameters[i]))
            scale.append([1]*len(building_parameters[i]))
            shift.append([0]*len(building_parameters[i]))
        
        for i in range(len(joint_parameters)):
            distribution = self.choose_distribution(buildings, joint_parameters[joint_names[i]], mode='auto')
            JointParameter = uv.Variable(joint_names[i], distribution)
            Q_joint.add(JointParameter)
            scale0 = distribution.get_bounds()[1] - distribution.get_bounds()[0]
            shift0 = distribution.get_bounds()[0]
            for j in range(len(building_parameters)):
                param_name = joint_parameters[joint_names[i]][j]
                if param_name != '-':
                    idx = buildings[j].Q.param_names().index(param_name)
                    indices[j][idx] = i
                    
                    lower_bound = buildings[j].Q.get_bounds()[idx][0]
                    higher_bound = buildings[j].Q.get_bounds()[idx][1]
                    scale[j][idx] = (higher_bound - lower_bound) / scale0
                    shift[j][idx] = lower_bound - scale0*shift0
                    building_parameters[j].remove(param_name)

        for i in range(len(building_parameters)):
            for j in range(len(building_parameters[i])):
                Parameter = uv.Variable(building_parameters[i][j], buildings[i].Q.params[building_parameters[i][j]])    
                Q_joint.add(Parameter)

        for i in range(len(building_parameters)):
            for j in range(len(building_parameters[i])):
                parameter = building_parameters[i][j]
                idx = Q_joint.param_names().index(parameter)
                original_index = buildings[i].Q.param_names().index(parameter)
                indices[i][original_index]  = idx   
        
        return Q_joint, indices, scale, shift
    
    def get_logprob(self, q):
        """ Compute log-probability for MCMC

            Parameters
            ----------
            q : array_like
                Parameter values

            Returns
            -------
            logprob : float
                Log-probability of the parameter values """
        
        if self.Q_ == 'default':
            logprob = self.loglikelihood(q, self.y_m) + self.logprior(q)
            return logprob
        else:
            logprob = self.loglikelihood(q, self.y_m) + self.Q_.logpdf(q)
            return logprob
        
    def logprior(self, q):
        """ Compute log-prior probability

            Parameters
            ----------
            q : array_like
                Parameter values

            Returns
            -------
            logpr : float
                Log-prior probability of the parameter values """
        
        logpr = self.Q.logpdf(q.reshape(-1,1))
        return logpr
    
    def loglikelihood(self, q, y_m):
        """ Compute log-likelihood of observed data

            Parameters
            ----------
            q : array_like
                Parameter values

            y_m : array_like
                Measured data

            Returns
            -------
            logp : float
                Log-likelihood of the observed data given the parameter values """
        
        q_df = pd.DataFrame(q.reshape(1,-1), columns=self.Q.param_names())
        predictions = np.zeros((1, 0))
        for i in range(len(self.models)):
            q_model = q[self.indices[i]] * self.scale[i] + self.shift[i]
            q_df = pd.DataFrame(q_model.reshape(1,-1), columns=self.models[i].Q.param_names())
            y_model = self.models[i].predict(q_df)
            predictions = np.concatenate((predictions, y_model), axis=1)
        d = y_m - predictions
        d = d.transpose()
        logp = self.E.logpdf(d)
        return logp
    
    def update(self, y_list, sigmas, nwalkers=150, nburn=100, niter=350, Q_='default', plot_samples=False):
        """ Perform MCMC update with observed data

            Parameters
            ----------
            y_list : list of pandas.DataFrame
                List of observed data DataFrames for each building

            sigmas : list of pandas.DataFrame
                List of standard deviation DataFrames for each building

            nwalkers : int, optionnal
                Number of MCMC walkers, by default 150

            nburn : int, optional
                Number of burn-in steps, by default 100

            niter : int, optional
                Number of MCMC iterations, by default 350

            Q_ : str or VariableSet, optional
                VariableSet object defining the prior distribution of parameters, by default 'default'

            plot_samples : bool, optional
                If True, plot MCMC samples, by default False """
        
        self.Q_ = Q_
        y_m = np.zeros((1, 0))
        for i in range(len(y_list)):
            y_building = y_list[i].to_numpy()
            y_m = np.concatenate((y_m, y_building), axis=1)
        self.y_m = y_m
        num_param = self.Q.num_params()
        self.E = self.generate_joint_stdrn_simparamset(sigmas)


        if Q_ == 'default':
            #logprob = lambda q: self.loglikelihood(q, y_m) + self.logprior(q)
            p0 = self.Q.sample(nwalkers)
        else:
            #logprob = lambda q: self.loglikelihood(q, y_m) + Q_.logpdf(q)
            p0 = Q_.sample(nwalkers)
        self.p0 = p0

        
        print('MCMC creating')
        sampler = emcee.EnsembleSampler(nwalkers, num_param, self.get_logprob) #pool=pool
        start_time = time.time()

        print('Burning period')
        state = sampler.run_mcmc(p0, nburn, progress = True)
        sampler.reset()

        print('MCMC running')
        sampler.run_mcmc(state, niter, progress = True)
    
        print("--- %s seconds ---" % (time.time() - start_time))
        self.sampler = sampler

    def get_mean_and_var_by_model(self, scaled=True):
        """ Get mean and variance of posterior for each building
        
            Parameters
            ----------
            scaled : bool, optional
                If True, return scaled mean and variance, by default True
                
            Returns
            -------
            mean_dfs : list of pandas.DataFrame
                List of DataFrames containing the mean of the posterior for each building
                
            varance_dfs : list of pandas.DataFrame
                List of DataFrames containing the variance of the posterior for each building """
        
        sampler = self.sampler
        post_samples = sampler.get_chain(flat=True)
        mean_dfs = []
        varance_dfs = []
        for i in range(len(self.models)):
            model_samples = post_samples[:, self.indices[i]]
            means = np.mean(model_samples, axis=0)
            variances = np.var(model_samples, axis=0)

            if scaled:
                means = means * self.scale[i] + self.shift[i]
                variances = variances * self.scale[i]

            means_df = pd.DataFrame(means.reshape(1,-1), columns=self.models[i].Q.param_names())
            variances_df = pd.DataFrame(variances.reshape(1,-1), columns=self.models[i].Q.param_names())
            mean_dfs.append(means_df)
            varance_dfs.append(variances_df)
        return mean_dfs, varance_dfs

    def get_MAP_by_model(self, scaled=True):
        """ Get maximum a posteriori estimate for each building
            
            Parameters
            ----------
            scaled : bool, optional
                If True, return scaled MAP estimate, by default True
                
            Returns
            -------
            map_dfs : list of pandas.DataFrame
                List of DataFrames containing the MAP estimate for each building """
        sampler = self.sampler
        post_samples = sampler.get_chain(flat=True)

        map_dfs = []
        for i in range(len(self.models)):
            map_estimate = np.zeros((1, len(self.indices[i])))
            model_samples = post_samples[:, self.indices[i]]
            for p in range(len(self.indices[i])):
                p_estimate = utils.estimate_maxima(model_samples[:,p])
                map_estimate[0,p] = p_estimate
            if scaled:
                map_estimate = map_estimate * self.scale[i] + self.shift[i]
            map_df = pd.DataFrame(map_estimate, columns=self.models[i].Q.param_names())
            map_dfs.append(map_df)
        return map_dfs        
        
    def get_MAP(self): # maximum a posterior estimate
        """ Get maximum a posteriori estimate of the joint variable set

            Returns
            -------
            map_df : pandas.DataFrame
                DataFrame containing the maximum a posteriori estimate of the joint variable set """
        
        sampler = self.sampler
        post_samples = sampler.get_chain(flat=True)
        
        map_estimate = np.zeros((1, post_samples.shape[1]))
        for p in range(post_samples.shape[1]):
            p_estimate = utils.estimate_maxima(post_samples[:,p])
            map_estimate[0,p] = p_estimate

        map_df = pd.DataFrame(map_estimate, columns=self.Q.param_names())
        return map_df
    
    def get_data_of_model(self, model_index):
        """ Get MAP, mean, variance, and samples for a specific building

            Parameters
            ----------
            model_index : int
                Index of the building model

            Returns
            -------
            map_df : pandas.DataFrame
                DataFrame containing the MAP estimate for the building
                
            mean_df : pandas.DataFrame
                DataFrame containing the mean of the posterior for the building
                
            var_df : pandas.DataFrame
                DataFrame containing the variance of the posterior for the building 
            
            sample_df : pandas.DataFrame
                DataFrame containing the samples from the posterior for the building"""
        
        maps = self.get_MAP_by_model()
        means, vars = self.get_mean_and_var_by_model()
        samples = self.get_posterior_samples_by_model()
        map_df = maps[model_index]
        mean_df = means[model_index]
        var_df = vars[model_index]
        sample_df = samples[model_index]
        return map_df, mean_df, var_df, sample_df
    
    def get_posterior_samples_by_model(self, scaled=True):
        """ Get posterior samples for each building

            Parameters
            ----------
            scaled : bool, optional
                If True, return scaled samples, by default True

            Returns
            -------
            sample_dfs : list of pandas.DataFrame
                List of DataFrames containing the samples from the posterior for each building"""
        
        sampler = self.sampler
        post_samples = sampler.get_chain(flat=True)
        sample_dfs = []
        for i in range(len(self.models)):
            model_samples = post_samples[:, self.indices[i]]
            if scaled:
                model_samples = model_samples * self.scale[i] + self.shift[i]
            sample_dfs.append(pd.DataFrame(model_samples, columns=self.models[i].Q.param_names()))
        return sample_dfs
    
    def get_posterior_samples(self):
        """ Get posterior samples for joint parameters

            Returns
            -------
            post_samples_df : pandas.DataFrame
                DataFrame containing the samples from the posterior parameter distribution """
        
        sampler = self.sampler
        post_samples = sampler.get_chain(flat=True)
        post_samples_df = pd.DataFrame(post_samples, columns=self.Q.param_names())
        return post_samples_df


import numpy as np
import matplotlib.pyplot as plt

import seaborn as sns
from IPython.display import display

from IPython.core.display import HTML

def pretty_print_df(df):
    """Display a pandas DataFrame as nicely formatted HTML in a notebook.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to display.

        Returns
        -------
        None """
    
    display(HTML(df.to_html()))

def print_tree_with_names(tree, column_names, indent):
    """ Convert a scikit-learn tree into an indented Python-like if/else representation.

        Parameters
        ----------
        tree : sklearn.tree._tree.Tree
            Fitted decision tree structure (e.g. from GradientBoosting or DecisionTree).
        column_names : list of str
            Feature names corresponding to tree.feature indices.
        indent : str
            String used for one indentation level (e.g. four spaces).

        Returns
        -------
        list of str
            Lines of code representing the tree as nested if/else statements.
    """
    left = tree.children_left
    right = tree.children_right
    threshold = tree.threshold
    features = [column_names[i] for i in tree.feature]
    value = tree.value
    depth = 0

    def get_indent(depth, indent):
        s = ""
        for j in range(depth):
            s += indent
        return s

    def recurse(left, right, threshold, features, node, depth):
        depth += 1
        text = []
        line_indent = get_indent(depth, indent)
        if (threshold[node] != -2):
            text += [line_indent + 'if row["%s"] <= %f:' % (features[node], threshold[node])]
            if left[node] != -1:
                text += recurse(left, right, threshold, features, left[node], depth)
            text += [line_indent + "else:"]
            if right[node] != -1:
                text += recurse(left, right, threshold, features, right[node], depth)
        else:
            text += [line_indent + 'score += %f' % value[node][0][0]]
        return text

    return recurse(left, right, threshold, features, 0, 1)


def plot_distribution(df, cols, outlier=0.01, x=1, y=1, bins=100, logscale=False):
    """Plot histograms for multiple columns with optional outlier filtering.

        Parameters
        ----------
        df : pandas.DataFrame
            Source DataFrame.
        cols : list of str
            Column names to plot.
        outlier : {'stddev', float} or float, optional
            If 'stddev', remove rows with values beyond 3 standard deviations.
            If float between 0 and 1, clip data to central (1-2*outlier) fraction.
            If 0 or False, no outlier filtering is applied.
        x : int, optional
            Number of subplot columns (auto-adjusted if too small).
        y : int, optional
            Number of subplot rows (auto-adjusted if too small).
        bins : int, optional
            Number of histogram bins.
        logscale : bool, optional
            If True, y-axis is logarithmic.

        Returns
        -------
        None """
    
    tmp = df[cols].copy()
    tmp.dropna(inplace=True)

        # outlier=='stddev': remove all rows with outliers beyond 3 times stddev
    if outlier == 'stdddev':
        tmp = tmp[tmp.apply(lambda x: np.abs(x - x.mean()) / x.std() < 3).all(axis=1)]

    if (x * y < len(cols)):
        x = int(np.floor(np.sqrt(len(cols))))
        y = int(np.ceil(len(cols) / x))

    sns.set(font_scale=0.8)
    sns.set_style("dark")

    fig, axes = plt.subplots(y, x, figsize=(10, 3))
    if (x * y > 1):
        ax = axes.ravel()
    else:
        ax = []
        ax.append(axes)
    i = 0
    for column in tmp.columns:
        if outlier != 'stddev' and outlier == outlier and outlier > 0 and outlier < 1:
            # filter by fraction
            outlier = round(outlier, 2)
            xmin = tmp[column].describe(percentiles=[outlier])[str(int(100 * outlier)) + '%']
            xmax = tmp[column].describe(percentiles=[1 - outlier])[str(int(100 * (1 - outlier))) + '%']
            ttmp = tmp[(tmp[column] > xmin) & (tmp[column] < xmax)]
        else:
            ttmp = tmp

        _, _bins = np.histogram(ttmp[column], bins=bins)
        ax[i].hist(ttmp[column], bins=_bins)
        # ax[i].set_yticks(())
        ax[i].set_title(column)
        if logscale:
            ax[i].set_yscale('log')
        i += 1

    fig.tight_layout()
    plt.show()


def normalize(df, cols, verbose=False):
    """ Normalize selected columns to [0, 1] using robust percentile-based scaling.

        For each variable, uses the 1st and 99th percentiles extended by 5% of their
        range, clipped to the observed min/max, then linearly maps to [0, 1].

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame containing variables to normalize.
        cols : list of str
            Column names to normalize.
        verbose : bool, optional
            If True, print scaling info for each variable.

        Returns
        -------
        list of str
            Names of the newly created normalized columns (original name + '_n'). """
    
    for var in cols:
        low = df[var].describe(percentiles=[0.01])['1%']
        high = df[var].describe(percentiles=[0.99])['99%']
        diff = high - low
        low = max(low - diff * 0.05, df[var].describe()['min'])
        high = min(high + diff * 0.05, df[var].describe()['max'])
        if (verbose):
            print(var, low, high)
            print(df[var].describe(percentiles=[0.01, 0.99]))
        df[var + '_n'] = df[var].apply(lambda x: x if x != x else min(1, max(0, (x - low) / (high - low))))
    return [x + '_n' for x in cols]

def corrplot(df, outlier=True, verbose=False, fixedscale=True, annot=True):
    """ Plot a correlation heatmap for a DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame
            Input data.
        outlier : bool, optional
            If True, apply percentile-based clipping before computing correlation.
        verbose : bool, optional
            If True, print scaling information for each variable when outlier is True.
        fixedscale : bool, optional
            If True, fix color scale to [-1, 1]; otherwise use data-driven limits.
        annot : bool, optional
            If True, annotate cells with correlation values.

        Returns
        -------
        None """
    
    tmp = df.copy()
    if outlier == True:
        for var in df.columns:
            low = df[var].describe(percentiles=[0.01])['1%']
            high = df[var].describe(percentiles=[0.99])['99%']
            diff = high - low
            low = max(low - diff * 0.05, df[var].describe()['min'])
            high = min(high + diff * 0.05, df[var].describe()['max'])
            if (verbose):
                print(var, low, high)
                print(df[var].describe(percentiles=[0.01, 0.99]))
            tmp[var] = df[var].apply(lambda x: x if x != x else min(1, max(0, (x - low) / (high - low))))

    corr = df.corr()
    _, dim = df.shape
    print('Minimum correlation:', min(corr.min()), 'Maximum correlation:', max(corr.max()))
    plt.figure(figsize=(dim, dim / 2))
    cmap = sns.diverging_palette(220, 5, as_cmap=True)
    if (fixedscale):
        sns.heatmap(corr,
                    cmap=cmap,
                    annot=annot,
                    vmax=1, vmin=-1,
                    xticklabels=corr.columns.values,
                    yticklabels=corr.columns.values)
    else:
        sns.heatmap(corr,
                    cmap=cmap,
                    annot=annot,
                    xticklabels=corr.columns.values,
                    yticklabels=corr.columns.values)
    plt.show()


def scatterplot(df, col1, col2, outlier=True, hue=None, threshold=None, regressionline=False):
    """ Scatter or regression plot of two columns with optional clipping and threshold line.

        Parameters
        ----------
        df : pandas.DataFrame
            Input data.
        col1 : str
            Column name for x-axis.
        col2 : str
            Column name for y-axis.
        outlier : bool, optional
            If True, limit axes to [1%, 99%] percentiles of each variable.
        hue : str or None, optional
            Column name for hue grouping in scatterplot.
        threshold : float or None, optional
            Optional vertical threshold line at x = threshold.
        regressionline : bool, optional
            If True, use seaborn.regplot with fitted regression line;
            otherwise use seaborn.scatterplot.

        Returns
        -------
        None """
    
    plt.figure(figsize=(10, 10))
    if regressionline:
        sp = sns.regplot(x=col1, y=col2, data=df, fit_reg=True, line_kws={'color': 'r'})
    else:
        sp = sns.scatterplot(x=col1, y=col2, hue=hue, data=df)
    if outlier == True:
        sp.set(xlim=(df[col1].describe(percentiles=[0.01])['1%'], df[col1].describe(percentiles=[0.99])['99%']))
        sp.set(ylim=(df[col2].describe(percentiles=[0.01])['1%'], df[col2].describe(percentiles=[0.99])['99%']))
        print(col1, df[col1].describe(percentiles=[0.01])['1%'], df[col1].describe(percentiles=[0.99])['99%'])
        print(col2, df[col2].describe(percentiles=[0.01])['1%'], df[col2].describe(percentiles=[0.99])['99%'])
    if threshold != None:
        plt.vlines(threshold, plt.gca().get_ylim()[0], plt.gca().get_ylim()[1])
    plt.show()
